/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#include <linux/slab.h> // kmalloc, kfree
#include <linux/module.h> // kmalloc, kfree, EXPORT_SYMBOL
#include <linux/interrupt.h>
#include <linux/sched.h>
#ifdef USERMODE_TEST
#include <linux/compatmac.h> // copy_from_user, copy_to_user
#endif
#include "linux_ver.h" // KMEM_CACHE, KMEM_ATOMIC
#include "sal/libc.h"
#include "sal/mem.h"
#include "sal/debug.h"

/* support for userlevel test application */
// #undef USERLEVEL_TEST

#ifdef DEBUG
    #ifdef DEBUG_LEVEL
        #ifndef DEBUGLEVEL_SAL_MEM
        #define DEBUGLEVEL_SAL_MEM DEBUG_LEVEL
        #endif
    #else
    #define DEBUGLEVEL_SAL_MEM 1
    #endif
#endif
#ifdef DEBUG
#define dbgl_salmem(l, x...) do { if ( l <= DEBUGLEVEL_SAL_MEM ) { sal_debug_print("MEM|%d|%s|",l, __FUNCTION__); sal_debug_println(x);}} while(0)
#else
#define dbgl_salmem(l, x...) do {} while(0)
#endif

#ifdef DEBUGLEVEL_SAL_MEM
#if DEBUGLEVEL_SAL_MEM > 4
#define MEM_DEBUG 
#else
#undef MEM_DEBUG
#endif
#endif

#ifdef MEM_DEBUG
#define MAX_NUM_MEM_BLOCKS 2048
#define MAX_NUM_MEM_SIZES  64

typedef struct
{
    void * addr;
    unsigned int size;
} mem_block_entry;

typedef struct
{
    int      count;
    int     accum_count;
    unsigned int  size;
} mem_block_counter;

static mem_block_entry v_mem_blocks[MAX_NUM_MEM_BLOCKS];
static mem_block_counter v_mem_counters[MAX_NUM_MEM_SIZES];
static long v_mem_total_allocated = 0;
static spinlock_t v_mem_block_lock = SPIN_LOCK_UNLOCKED;

static void set_mem_block_size_adjust(unsigned int size, int adj)
{
    int i;
    for(i=0;i<MAX_NUM_MEM_SIZES;i++) {
        if (v_mem_counters[i].size==0) { /* new entry */
            v_mem_counters[i].size = size; 
        } 
        if (v_mem_counters[i].size == size) {
            v_mem_counters[i].count += adj; 
            if (adj>0)
                v_mem_counters[i].accum_count +=adj;
            break;
        }
    }
    sal_assert(i<MAX_NUM_MEM_SIZES);
}
#endif

NDAS_SAL_API
void *sal_malloc(unsigned int size)
{
#ifdef MEM_DEBUG
    int i;
    unsigned long flags;
#endif
    void* ptr = kmalloc(size, GFP_ATOMIC);
    if (ptr==NULL) {
        printk("ndas: failed to allocate memory size=%d",size);
        return ptr;
    }
    dbgl_salmem(9, "allocated size=%d",size);
#ifdef MEM_DEBUG        
    spin_lock_irqsave(&v_mem_block_lock, flags);
    for (i = 0; i < MAX_NUM_MEM_BLOCKS; i++)
    {
        if (v_mem_blocks[i].addr == 0)
        {
            // found no_pending_dpc entry; use it 
            v_mem_blocks[i].addr = ptr;
            v_mem_blocks[i].size = size;
            set_mem_block_size_adjust(size, 1);
            break;
        }
    }
    v_mem_total_allocated += size;    
    spin_unlock_irqrestore(&v_mem_block_lock, flags);
    if (i >= MAX_NUM_MEM_BLOCKS) {
        dbgl_salmem(1,"Requested memory %d failed\n", size);
//        sal_mem_display(2);
    }
    sal_assert(i < MAX_NUM_MEM_BLOCKS);
#endif
    return ptr;
}
EXPORT_SYMBOL(sal_malloc);

NDAS_SAL_API
void sal_free(void *ptr)
{
#ifdef MEM_DEBUG    
    int i;
    unsigned long flags;
    spin_lock_irqsave(&v_mem_block_lock, flags);
    for (i = 0; i < MAX_NUM_MEM_BLOCKS; i++)
    {
        if (v_mem_blocks[i].addr == ptr)
        {
            // found block being released 
            set_mem_block_size_adjust(v_mem_blocks[i].size, -1);
            v_mem_total_allocated -= v_mem_blocks[i].size;
            v_mem_blocks[i].addr = 0;
            v_mem_blocks[i].size = 0;
            break;
        }
    }
    spin_unlock_irqrestore(&v_mem_block_lock, flags);
    if (i >= MAX_NUM_MEM_BLOCKS) {
        dbgl_salmem(1,"Failed find memory to free %x\n", ptr);
//        sal_mem_display(2);
    }
    sal_assert(i < MAX_NUM_MEM_BLOCKS);
#endif
    kfree(ptr);
}
EXPORT_SYMBOL(sal_free);

NDAS_SAL_API
void sal_mem_display(int verbose)
{
#ifdef MEM_DEBUG    
    int i,j;
    sal_debug_print("\nSize     Freq.     Accum.");
    for (i = 0; i < MAX_NUM_MEM_SIZES; i++)
    {
        if (v_mem_counters[i].size == 0) 
            break;
        sal_debug_print("\n%d    %d     %d", 
            v_mem_counters[i].size, v_mem_counters[i].count, v_mem_counters[i].accum_count);
    }
    if (verbose>=1) {
        sal_debug_print("\nAddress            Size");
        for (i = 0; i < MAX_NUM_MEM_BLOCKS; i++) {
            if (v_mem_blocks[i].addr != 0) {
                sal_debug_print("\n%p     %d",v_mem_blocks[i].addr, v_mem_blocks[i].size);
                if (verbose >=2) {
                    for(j=0;j<v_mem_blocks[i].size;j+=4) {
                        if (j%128==0) {
                            sal_debug_print("\n    ");
                        }
                        sal_debug_print("%08x ", *(((unsigned int*)v_mem_blocks[i].addr)+j));
                    }
                }
            }
        }
    }
    sal_debug_print("\nTotal memory allocated: %d\n", v_mem_total_allocated);
#else
    sal_debug_print("MEM_DEBUG flags is not turned on\n");
#endif        
}

NDAS_SAL_API void* sal_create_mem_pool(char* name, int size)
{
    return (void*)kmem_cache_create(name, size, 0, 0, NULL, NULL);
}
EXPORT_SYMBOL(sal_create_mem_pool);

NDAS_SAL_API int sal_destroy_mem_pool(void* pool)
{
    kmem_cache_destroy((KMEM_CACHE*)pool);
    return 0;
}
EXPORT_SYMBOL(sal_destroy_mem_pool);

NDAS_SAL_API void* sal_alloc_from_pool(void* pool, int size)
{
    return kmem_cache_alloc((KMEM_CACHE*)pool, KMEM_ATOMIC);
#if 0
    void* ptr;
    if (in_interrupt()) {
        ptr = kmem_cache_alloc((KMEM_CACHE*)pool, KMEM_ATOMIC);
    } else {
        ptr = kmem_cache_alloc((KMEM_CACHE*)pool, SLAB_KERNEL|GFP_KERNEL);
    }
    return ptr;
#endif    
}
EXPORT_SYMBOL(sal_alloc_from_pool);

NDAS_SAL_API void sal_free_from_pool(void* pool, void* ptr)
{
    kmem_cache_free((KMEM_CACHE*)pool, ptr);
}
EXPORT_SYMBOL(sal_free_from_pool);

