/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#include <linux/version.h> // LINUX_VERSION
#include <linux/module.h> // EXPORT_SYMBOL
#include <linux/spinlock.h> // spinklock_t
#include <linux/sched.h> // jiffies
#include <asm/semaphore.h> // struct semaphore
#include <asm/atomic.h> // atomic
#include <linux/mm.h> //
#include <linux/slab.h> // 2.4 kmalloc/kfree
#include <linux/interrupt.h>

#include "sal/sync.h"
#include "sal/thread.h"
#include "sal/time.h"
#include "sal/mem.h"
#include "sal/debug.h"
#include "sal/types.h"
#include "linux_ver.h" // SAL_HZ

#ifdef DEBUG
#ifndef DEBUG_LEVEL_SAL_SYNC
#ifdef DEBUG_LEVEL
#define DEBUG_LEVEL_SAL_SYNC DEBUG_LEVEL
#else
#define DEBUG_LEVEL_SAL_SYNC 1
#endif
#endif
#define dbgl_salsync(l,x...) do {\
    if ( l <= DEBUG_LEVEL_SAL_SYNC ) {\
        sal_debug_print("SS|%d|%s|",l,__FUNCTION__); \
        sal_debug_println(x); \
    } \
} while(0) 
#else
#define dbgl_salsync(l,x...) do {} while(0)
#endif
#define SAL_MUTEX_MAGIC 0x9832abcd
#define SAL_SEMA_MAGIC 0x9832abcd

//#define DEBUG_TRACE_SEMAPHORE 1

#ifndef XPLAT_ENABLE_EMBED_LINUX_OPT
struct _sal_mutex {
#undef MAGIC 
#ifdef MAGIC
    int magic;
#endif
    spinlock_t    mutex;
    const char        *desc;
};

NDAS_SAL_API
sal_mutex
sal_mutex_create(const char *desc)
{
    struct _sal_mutex    *sm;

    if ((sm = kmalloc(sizeof (struct _sal_mutex), GFP_ATOMIC)) == NULL) {
        return NULL;
    }

    dbgl_salsync(5, "%p %s", sm, desc);
    
    sm->desc = desc;
#ifdef MAGIC
    sm->magic = SAL_MUTEX_MAGIC;
#endif
    spin_lock_init(&sm->mutex);
    return (sal_mutex)sm;
}
EXPORT_SYMBOL(sal_mutex_create);

NDAS_SAL_API
void
sal_mutex_destroy(sal_mutex m)
{
    struct _sal_mutex    *sm;
    sm = (struct _sal_mutex*) m;  
    sal_assert(m!=SAL_INVALID_MUTEX);
#ifdef    MAGIC
    sal_assert(m->magic == SAL_MUTEX_MAGIC);
#endif
#ifdef DEBUG
    dbgl_salsync(5, "%s %p",sm->desc, sm);    
    sal_assert( !spin_is_locked(&sm->mutex) );
#endif
    // check if locked and print debugging info
    kfree(sm);
}
EXPORT_SYMBOL(sal_mutex_destroy);

/* return -1 for timeout, 0 for successful mutex taking */
NDAS_SAL_API
void
sal_mutex_take(sal_mutex m, unsigned long* flags)
{
    struct _sal_mutex    *sm = (struct _sal_mutex*) m;
    
    sal_assert(m!=SAL_INVALID_MUTEX);
#ifdef MAGIC
    sal_assert(m->magic == SAL_MUTEX_MAGIC);
#endif

    dbgl_salsync(5, "ing(%s)",sm->desc);
    spin_lock_irqsave(&sm->mutex,*flags);
    dbgl_salsync(5, "ed(%s)",sm->desc);
}
EXPORT_SYMBOL(sal_mutex_take);

NDAS_SAL_API
void
sal_mutex_give(sal_mutex m, unsigned long* flags)
{
    struct _sal_mutex    *sm = (struct _sal_mutex*) m;
    while(0) {sm = sm;} /* sm may not used in some kernel */    

    sal_assert(m!=SAL_INVALID_MUTEX);
#ifdef MAGIC
    sal_assert(m->magic == SAL_MUTEX_MAGIC);
#endif
    dbgl_salsync(5, "ing (%s)",(sm->desc));
    spin_unlock_irqrestore(&sm->mutex, *flags);
    dbgl_salsync(5, "ed (%s)",(sm->desc));
}
EXPORT_SYMBOL(sal_mutex_give);
#endif /* XPLAT_ENABLE_EMBED_LINUX_OPT */

struct _sal_semaphore {
    spinlock_t lock;
    atomic_t val;
    wait_queue_head_t wait;
//    struct timer_list timer;
    char* desc;
};

NDAS_SAL_API
sal_semaphore
sal_semaphore_create(const char *desc, int max_count, int initial_count)
{
    struct _sal_semaphore        *s;

    /* Currently support binary semaphore only */ 
    sal_assert(max_count==1);
    sal_assert(initial_count>=0 && initial_count<=max_count);

    if ((s = kmalloc(sizeof(struct _sal_semaphore),GFP_ATOMIC)) != 0) {
        init_waitqueue_head (&s->wait);
        atomic_set(&s->val, initial_count);
//        s->val = initial_count;
        spin_lock_init(&s->lock); 
//	init_timer(&s->timer);
//	s->timer.function = _sal_semaphore_timeout;
//	s->timer.data = (unsigned long)s;
        s->desc = (char*)desc;
    }
    sal_assert(s != SAL_INVALID_SEMAPHORE);

    return (sal_semaphore) s;
}
EXPORT_SYMBOL(sal_semaphore_create);

NDAS_SAL_API
void
sal_semaphore_destroy(sal_semaphore b)
{
    struct _sal_semaphore        *s = (struct _sal_semaphore *) b;
    sal_assert(b!=SAL_INVALID_SEMAPHORE);
    kfree(s);
}
EXPORT_SYMBOL(sal_semaphore_destroy);

/* Return -1 for timeout, 0 for successful semaphore taking */
NDAS_SAL_API
int
sal_semaphore_take(sal_semaphore b, sal_tick tick)
{
    struct _sal_semaphore   * s = (struct _sal_semaphore *) b;
    unsigned long flags;
    sal_tick prev_tick;
    long remain;
    sal_assert(b != SAL_INVALID_SEMAPHORE);
    sal_assert(b != (sal_semaphore) 0xc);
    sal_assert(tick == SAL_SYNC_NOWAIT  || !in_interrupt());
    
    spin_lock_irqsave(&s->lock, flags);    
    if (atomic_read(&s->val)>0) {
        atomic_dec(&s->val);
        spin_unlock_irqrestore(&s->lock, flags);
        return 0;
    } else if (tick == SAL_SYNC_NOWAIT) {
        spin_unlock_irqrestore(&s->lock, flags);
        return -1;
    }
    spin_unlock_irqrestore(&s->lock, flags);
    
    if (tick == SAL_SYNC_FOREVER) {
        while(1) {
            /* currently ndascore does not handle interrupted wait. Just block interrupt */
            /* wait_event_interruptible(s->wait, atomic_read(&s->val)>0); */
            wait_event(s->wait, atomic_read(&s->val)>0);
            spin_lock_irqsave(&s->lock, flags);
            if (atomic_read(&s->val)>0) {
                atomic_dec(&s->val);
                spin_unlock_irqrestore(&s->lock, flags);
                return 0;
            }
            spin_unlock_irqrestore(&s->lock, flags);
            // Not my event. sleep more
        }
    } 

    remain = tick;
    while(1) {
        prev_tick = sal_get_tick();
#if LINUX_VERSION_25_ABOVE
        /* wait_event_interruptible_timeout(s->wait, atomic_read(&s->val)> 0, remain); */
        wait_event_timeout(s->wait, atomic_read(&s->val)> 0, remain);
#else
	    /* Wait till resource is ready or time is over */
    	spin_lock_irqsave(&s->lock, flags);
    	if (atomic_read(&s->val)<=0){ 
    		spin_unlock_irqrestore(&s->lock, flags);
    		interruptible_sleep_on_timeout(&s->wait, HZ/50);
    	} else {
    	    spin_unlock_irqrestore(&s->lock, flags);
    	}
#endif
        remain -= sal_tick_sub(sal_get_tick(),prev_tick);
        spin_lock_irqsave(&s->lock, flags);
        if (atomic_read(&s->val)>0) {
            atomic_dec(&s->val);
            spin_unlock_irqrestore(&s->lock, flags);          
            return 0;
        }
        spin_unlock_irqrestore(&s->lock, flags);
        if (remain<=0) {
            return -1;
        }
    }
}
EXPORT_SYMBOL(sal_semaphore_take);

NDAS_SAL_API int        sal_semaphore_give(sal_semaphore b)
{
    struct _sal_semaphore   * s = (struct _sal_semaphore *) b;
    unsigned long flags;
    spin_lock_irqsave(&s->lock, flags);
    if (waitqueue_active(&s->wait)) {
        /*  wake_up_interruptible(&s->wait); */
        wake_up(&s->wait);
    } 
    
    atomic_inc(&s->val);
    spin_unlock_irqrestore(&s->lock, flags);
    return 0;
}

EXPORT_SYMBOL(sal_semaphore_give);

NDAS_SAL_API int sal_semaphore_getvalue(sal_semaphore b)
{
    struct _sal_semaphore   * s = (struct _sal_semaphore *) b;
    return atomic_read(&s->val);
}
EXPORT_SYMBOL(sal_semaphore_getvalue);

